__all__ = ('MessageSearchQuery',)

from scarletio import RichAttributeErrorBaseType

from ...utils import DATETIME_FORMAT_CODE

from .constants import LIMIT_DEFAULT, OFFSET_DEFAULT, SLOP_DEFAULT
from .fields import (
    parse_after, parse_attachment_extensions, parse_attachment_names, parse_author_ids, parse_author_types,
    parse_before, parse_channel_ids, parse_content, parse_embed_providers, parse_embed_types, parse_has_types,
    parse_include_nsfw_channels, parse_limit, parse_mentioned_everyone, parse_mentioned_user_ids, parse_offset,
    parse_pinned, parse_slop, parse_sort_by, parse_sort_order, parse_url_host_names, put_after,
    put_attachment_extensions, put_attachment_names, put_author_ids, put_author_types, put_before, put_channel_ids,
    put_content, put_embed_providers, put_embed_types, put_has_types, put_include_nsfw_channels, put_limit,
    put_mentioned_everyone, put_mentioned_user_ids, put_offset, put_pinned, put_slop, put_sort_by, put_sort_order,
    put_url_host_names, validate_after, validate_attachment_extensions, validate_attachment_names, validate_author_ids,
    validate_author_types, validate_before, validate_channel_ids, validate_content, validate_embed_providers,
    validate_embed_types, validate_has_types, validate_include_nsfw_channels, validate_limit,
    validate_mentioned_everyone, validate_mentioned_user_ids, validate_offset, validate_pinned, validate_slop,
    validate_sort_by, validate_sort_order, validate_url_host_names
)
from .preinstanced import MessageSearchSortByType, MessageSearchSortOrderType


SERIALISATION_SHIFT_AFTER = 0
SERIALISATION_SHIFT_ATTACHMENT_EXTENSIONS = 1
SERIALISATION_SHIFT_ATTACHMENT_NAMES = 2
SERIALISATION_SHIFT_AUTHOR_IDS = 3
SERIALISATION_SHIFT_AUTHOR_TYPES = 4
SERIALISATION_SHIFT_BEFORE = 5
SERIALISATION_SHIFT_CHANNEL_IDS = 6
SERIALISATION_SHIFT_CONTENT = 7
SERIALISATION_SHIFT_EMBED_PROVIDERS = 8
SERIALISATION_SHIFT_EMBED_TYPES = 9
SERIALISATION_SHIFT_HAS_TYPES = 10
SERIALISATION_SHIFT_INCLUDE_NSFW_CHANNELS = 11
SERIALISATION_SHIFT_LIMIT = 12
SERIALISATION_SHIFT_MENTIONED_EVERYONE = 13
SERIALISATION_SHIFT_MENTIONED_USER_IDS = 14
SERIALISATION_SHIFT_OFFSET = 15
SERIALISATION_SHIFT_PINNED = 16
SERIALISATION_SHIFT_SLOP = 17
SERIALISATION_SHIFT_SORT_BY = 18
SERIALISATION_SHIFT_SORT_ORDER = 19
SERIALISATION_SHIFT_URL_HOST_NAMES = 20


class MessageSearchQuery(RichAttributeErrorBaseType):
    """
    Message search query.
    
    Attributes
    ----------
    _serialisation_flags : `int`
        Flags used internally to know what to force serialise.
    
    after : `None | DateTime`
        Message creation time lower threshold.
    
    attachment_extensions : `None | tuple<str>`
        Whether the returned messages should have attachments with these extensions.
    
    attachment_names : `None | tuple<str>`
        Whether the returned messages should have attachments with these names.
    
    author_ids : `None | tuple<int>`
        Whether the returned messages should be created by these users defined by their identifiers.
    
    author_types : ``None | tuple<MessageSearchAuthorType>``
        Whether the returned messages should be created by these type of users.
    
    before : `None | DateTime`
        Message creation time upper threshold.
    
    channel_ids : `None | tuple<int>`
        Whether the returned messages should be created in these channels defined by their identifiers.
    
    content : `None | str`
        Content to search in the messages.
    
    embed_providers : `None | tuple<str>`
        Whether the returned messages should have embeds with these providers.
    
    embed_types : ``None | tuple<EmbedType>``
        Whether the returned messages should have embeds with these types.
    
    has_types : ``None | tuple<MessageSearchHasType>``
        What type of content should the returned messages have.
    
    include_nsfw_channels : `bool`
        Whether to include nsfw channels in the search.
    
    limit : `int`
        The amount of messages to return.
    
    mentioned_everyone : `bool`
        Whether the returned messages should be mentioning everyone.
    
    mentioned_user_ids : `None | tuple<int>`
        Whether the returned messages should be mentioned these users by their identifiers.
    
    offset : `int`
        Amount to offset the returned messages by.
    
    pinned : `bool`
        Whether the returned messages should be pinned.
    
    slop : `int`
        The maximal amount of words that can be in between each word from `content`.
    
    sort_by : ``MessageSearchSortByType``
        By what to sort the returned messages.
    
    sort_order : ``MessageSearchSortOrderType``
        How to order the returned messages. Depending on the `sort_by`, this field may not be respected.
    
    url_host_names : `None | tuple<str>`
        Whether the returned messages should contain urls with these host names.
    """
    __slots__ = (
        '_serialisation_flags', 'after', 'attachment_extensions', 'attachment_names', 'author_ids', 'author_types',
        'before', 'channel_ids', 'content', 'embed_providers', 'embed_types', 'has_types', 'include_nsfw_channels',
        'limit', 'mentioned_everyone', 'mentioned_user_ids', 'offset', 'pinned', 'slop', 'sort_by', 'sort_order',
        'url_host_names'
    )
    
    
    def __new__(
        cls,
        *,
        after = ...,
        attachment_extensions = ...,
        attachment_names = ...,
        author_ids = ...,
        author_types = ...,
        before = ...,
        channel_ids = ...,
        content = ...,
        embed_providers = ...,
        embed_types = ...,
        has_types = ...,
        include_nsfw_channels = ...,
        limit = ...,
        mentioned_everyone = ...,
        mentioned_user_ids = ...,
        offset = ...,
        pinned = ...,
        slop = ...,
        sort_by = ...,
        sort_order = ...,
        url_host_names = ...,
    ):
        """
        Creates a new messages search query.
        
        Parameters
        ----------
        after : `None | DateTime`, Optional (Keyword only)
            Message creation time lower threshold.
        
        attachment_extensions : `None | str | iterable<str>`, Optional (Keyword only)
            Whether the returned messages should have attachments with these extensions.
        
        attachment_names : `None | str | iterable<str>`, Optional (Keyword only)
            Whether the returned messages should have attachments with these names.
        
        author_ids : ``None | int | iterable<int> | UserBase | iterable<UserBase>``, Optional (Keyword only)
            Whether the returned messages should be created by these users defined by their identifiers.
        
        author_types : ``None | str | iterable<str> | MessageSearchAuthorType | iterable<MessageSearchAuthorType>`` \
                , Optional (Keyword only)
            Whether the returned messages should be created by these type of users.
        
        before : `None | DateTime`, Optional (Keyword only)
            Message creation time upper threshold.
        
        channel_ids : ``None | int | iterable<int> | Channel | iterable<Channel>``, Optional (Keyword only)
            Whether the returned messages should be created in these channels defined by their identifiers.
        
        content : `None | str`, Optional (Keyword only)
            Content to search in the messages.
        
        embed_providers : `None | str | iterable<str>`, Optional (Keyword only)
            Whether the returned messages should have embeds with these providers.
        
        embed_types : ``None | str | iterable<str> | EmbedType | iterable<EmbedType>``, Optional (Keyword only)
            Whether the returned messages should have embeds with these types.
        
        has_types : ``None | str | iterable<str> | MessageSearchHasType | iterable<MessageSearchHasType>`` \
                , Optional (Keyword only)
            What type of content should the returned messages have.
        
        include_nsfw_channels : `None | bool`, Optional (Keyword only)
            Whether to include nsfw channels in the search.
        
        limit : `None | int`, Optional (Keyword only)
            The amount of messages to return.
        
        mentioned_everyone : `None | bool`, Optional (Keyword only)
            Whether the returned messages should be mentioning everyone.
        
        mentioned_user_ids : ``None | int | iterable<int> | ClientUserBase | iterable<ClientUserBase>`` \
                , Optional (Keyword only)
            Whether the returned messages should be mentioned these users by their identifiers.
        
        offset : `None | int`, Optional (Keyword only)
            Amount to offset the returned messages by.
        
        pinned : `None | bool`, Optional (Keyword only)
            Whether the returned messages should be pinned.
        
        slop : `None | int`, Optional (Keyword only)
            The maximal amount of words that can be in between each word from `content`.
        
        sort_by : ``None | str | MessageSearchSortByType``, Optional (Keyword only)
            By what to sort the returned messages.
        
        sort_order : ``None | str | MessageSearchSortOrderType``, Optional (Keyword only)
            How to order the returned messages. Depending on the `sort_by`, this field may not be respected.
        
        url_host_names : `None | str | iterable<str>`, Optional (Keyword only)
            Whether the returned messages should contain urls with these host names.
        
        Raises
        ------
        ValueError
            - If a parameter's value is incorrect.
        TypeError
            - If a parameter's type is incorrect.
        """
        serialisation_flags = 0
        
        # after
        if (after is ...):
            after = None
        else:
            after = validate_after(after)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_AFTER)
        
        # attachment_extensions
        if (attachment_extensions is ...):
            attachment_extensions = None
        else:
            attachment_extensions = validate_attachment_extensions(attachment_extensions)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_ATTACHMENT_EXTENSIONS)
        
        # attachment_names
        if (attachment_names is ...):
            attachment_names = None
        else:
            attachment_names = validate_attachment_names(attachment_names)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_ATTACHMENT_NAMES)
        
        # author_ids
        if (author_ids is ...):
            author_ids = None
        else:
            author_ids = validate_author_ids(author_ids)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_AUTHOR_IDS)
        
        # author_types
        if (author_types is ...):
            author_types = None
        else:
            author_types = validate_author_types(author_types)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_AUTHOR_TYPES)
        
        # before
        if (before is ...):
            before = None
        else:
            before = validate_before(before)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_BEFORE)
        
        # channel_ids
        if (channel_ids is ...):
            channel_ids = None
        else:
            channel_ids = validate_channel_ids(channel_ids)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_CHANNEL_IDS)
        
        # content
        if (content is ...):
            content = None
        else:
            content = validate_content(content)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_CONTENT)
        
        # embed_providers
        if (embed_providers is ...):
            embed_providers = None
        else:
            embed_providers = validate_embed_providers(embed_providers)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_EMBED_PROVIDERS)
        
        # embed_types
        if (embed_types is ...):
            embed_types = None
        else:
            embed_types = validate_embed_types(embed_types)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_EMBED_TYPES)
        
        # has_types
        if (has_types is ...):
            has_types = None
        else:
            has_types = validate_has_types(has_types)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_HAS_TYPES)
        
        # include_nsfw_channels
        if (include_nsfw_channels is ...):
            include_nsfw_channels = False
        else:
            include_nsfw_channels = validate_include_nsfw_channels(include_nsfw_channels)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_INCLUDE_NSFW_CHANNELS)
        
        # limit
        if (limit is ...):
            limit = LIMIT_DEFAULT
        else:
            limit = validate_limit(limit)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_LIMIT)
        
        # mentioned_everyone
        if (mentioned_everyone is ...):
            mentioned_everyone = False
        else:
            mentioned_everyone = validate_mentioned_everyone(mentioned_everyone)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_MENTIONED_EVERYONE)
        
        # mentioned_user_ids
        if (mentioned_user_ids is ...):
            mentioned_user_ids = None
        else:
            mentioned_user_ids = validate_mentioned_user_ids(mentioned_user_ids)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_MENTIONED_USER_IDS)
        
        # offset
        if (offset is ...):
            offset = OFFSET_DEFAULT
        else:
            offset = validate_offset(offset)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_OFFSET)
        
        # pinned
        if (pinned is ...):
            pinned = False
        else:
            pinned = validate_pinned(pinned)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_PINNED)
        
        # slop
        if (slop is ...):
            slop = SLOP_DEFAULT
        else:
            slop = validate_slop(slop)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_SLOP)
        
        # sort_by
        if (sort_by is ...):
            sort_by = MessageSearchSortByType.creation
        else:
            sort_by = validate_sort_by(sort_by)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_SORT_BY)
        
        # sort_order
        if (sort_order is ...):
            sort_order = MessageSearchSortOrderType.descending
        else:
            sort_order = validate_sort_order(sort_order)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_SORT_ORDER)
        
        # url_host_names
        if (url_host_names is ...):
            url_host_names = None
        else:
            url_host_names = validate_url_host_names(url_host_names)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_URL_HOST_NAMES)
        
        # Creation
        self = object.__new__(cls)
        self._serialisation_flags = serialisation_flags
        self.after = after
        self.attachment_extensions = attachment_extensions
        self.attachment_names = attachment_names
        self.author_ids = author_ids
        self.author_types = author_types
        self.before = before
        self.channel_ids = channel_ids
        self.content = content
        self.embed_providers = embed_providers
        self.embed_types = embed_types
        self.has_types = has_types
        self.include_nsfw_channels = include_nsfw_channels
        self.limit = limit
        self.mentioned_everyone = mentioned_everyone
        self.mentioned_user_ids = mentioned_user_ids
        self.offset = offset
        self.pinned = pinned
        self.slop = slop
        self.sort_by = sort_by
        self.sort_order = sort_order
        self.url_host_names = url_host_names
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        serialisation_flags = self._serialisation_flags
        field_added = False
        
        repr_parts = ['<', type(self).__name__]
        
        # after
        if (serialisation_flags >> SERIALISATION_SHIFT_AFTER) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' after = ')
            after = self.after
            repr_parts.append(repr(None) if after is None else format(after, DATETIME_FORMAT_CODE))
        
        # attachment_extensions
        if (serialisation_flags >> SERIALISATION_SHIFT_ATTACHMENT_EXTENSIONS) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' attachment_extensions = ')
            repr_parts.append(repr(self.attachment_extensions))
        
        # attachment_names
        if (serialisation_flags >> SERIALISATION_SHIFT_ATTACHMENT_NAMES) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' attachment_names = ')
            repr_parts.append(repr(self.attachment_names))
        
        # author_ids
        if (serialisation_flags >> SERIALISATION_SHIFT_AUTHOR_IDS) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' author_ids = ')
            repr_parts.append(repr(self.author_ids))
        
        # author_types
        if (serialisation_flags >> SERIALISATION_SHIFT_AUTHOR_TYPES) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' author_types = ')
            repr_parts.append(repr(self.author_types))
        
        # before
        if (serialisation_flags >> SERIALISATION_SHIFT_BEFORE) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' before = ')
            before = self.before
            repr_parts.append(repr(None) if before is None else format(before, DATETIME_FORMAT_CODE))
        
        # channel_ids
        if (serialisation_flags >> SERIALISATION_SHIFT_CHANNEL_IDS) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' channel_ids = ')
            repr_parts.append(repr(self.channel_ids))
        
        # content
        if (serialisation_flags >> SERIALISATION_SHIFT_CONTENT) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' content = ')
            repr_parts.append(repr(self.content))
        
        # embed_providers
        if (serialisation_flags >> SERIALISATION_SHIFT_EMBED_PROVIDERS) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' embed_providers = ')
            repr_parts.append(repr(self.embed_providers))
        
        # embed_types
        if (serialisation_flags >> SERIALISATION_SHIFT_EMBED_TYPES) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' embed_types = ')
            repr_parts.append(repr(self.embed_types))
        
        # has_types
        if (serialisation_flags >> SERIALISATION_SHIFT_HAS_TYPES) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' has_types = ')
            repr_parts.append(repr(self.has_types))
        
        # include_nsfw_channels
        if (serialisation_flags >> SERIALISATION_SHIFT_INCLUDE_NSFW_CHANNELS) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' include_nsfw_channels = ')
            repr_parts.append(repr(self.include_nsfw_channels))
        
        # limit
        if (serialisation_flags >> SERIALISATION_SHIFT_LIMIT) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' limit = ')
            repr_parts.append(repr(self.limit))
        
        # mentioned_everyone
        if (serialisation_flags >> SERIALISATION_SHIFT_MENTIONED_EVERYONE) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' mentioned_everyone = ')
            repr_parts.append(repr(self.mentioned_everyone))
        
        # mentioned_user_ids
        if (serialisation_flags >> SERIALISATION_SHIFT_MENTIONED_USER_IDS) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' mentioned_user_ids = ')
            repr_parts.append(repr(self.mentioned_user_ids))
        
        # offset
        if (serialisation_flags >> SERIALISATION_SHIFT_OFFSET) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' offset = ')
            repr_parts.append(repr(self.offset))
        
        # pinned
        if (serialisation_flags >> SERIALISATION_SHIFT_PINNED) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' pinned = ')
            repr_parts.append(repr(self.pinned))
        
        # slop
        if (serialisation_flags >> SERIALISATION_SHIFT_SLOP) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' slop = ')
            repr_parts.append(repr(self.slop))
        
        # sort_by
        if (serialisation_flags >> SERIALISATION_SHIFT_SORT_BY) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' sort_by = ')
            repr_parts.append(repr(self.sort_by))
        
        # sort_order
        if (serialisation_flags >> SERIALISATION_SHIFT_SORT_ORDER) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' sort_order = ')
            repr_parts.append(repr(self.sort_order))
        
        # url_host_names
        if (serialisation_flags >> SERIALISATION_SHIFT_URL_HOST_NAMES) & 1:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' url_host_names = ')
            repr_parts.append(repr(self.url_host_names))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = self._serialisation_flags
        
        # after
        after = self.after
        if (after is not None):
            hash_value ^= hash(after)
        
        # attachment_extensions
        attachment_extensions = self.attachment_extensions
        if (attachment_extensions is not None):
            hash_value ^= hash(attachment_extensions)
        
        # attachment_names
        attachment_names = self.attachment_names
        if (attachment_names is not None):
            hash_value ^= hash(attachment_names)
        
        # author_ids
        author_ids = self.author_ids
        if (author_ids is not None):
            hash_value ^= hash(author_ids)
        
        # author_types
        author_types = self.author_types
        if (author_types is not None):
            hash_value ^= hash(author_types)
        
        # before
        before = self.before
        if (before is not None):
            hash_value ^= hash(before)
        
        # channel_ids
        channel_ids = self.channel_ids
        if (channel_ids is not None):
            hash_value ^= hash(channel_ids)
        
        # content
        content = self.content
        if (content is not None):
            hash_value ^= hash(content)
        
        # embed_providers
        embed_providers = self.embed_providers
        if (embed_providers is not None):
            hash_value ^= hash(embed_providers)
        
        # embed_types
        embed_types = self.embed_types
        if (embed_types is not None):
            hash_value ^= hash(embed_types)
        
        # has_types
        has_types = self.has_types
        if (has_types is not None):
            hash_value ^= hash(has_types)
        
        # include_nsfw_channels
        hash_value ^= self.include_nsfw_channels << 32
        
        # limit
        hash_value ^= self.limit << 33
        
        # mentioned_everyone
        hash_value ^= self.mentioned_everyone << 35
        
        # mentioned_user_ids
        mentioned_user_ids = self.mentioned_user_ids
        if (mentioned_user_ids is not None):
            hash_value ^= hash(mentioned_user_ids)
        
        # offset
        hash_value ^= self.offset << 36
        
        # pinned
        hash_value ^= self.pinned << 41
        
        # slop
        hash_value ^= self.slop << 42
        
        # sort_by
        hash_value ^= hash(self.sort_by)
        
        # sort_order
        hash_value ^= hash(self.sort_order)
        
        # url_host_names
        url_host_names = self.url_host_names
        if (url_host_names is not None):
            hash_value ^= hash(url_host_names)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # _serialisation_flags
        if self._serialisation_flags != other._serialisation_flags:
            return False
        
        # after
        if self.after != other.after:
            return False
        
        # attachment_extensions
        if self.attachment_extensions != other.attachment_extensions:
            return False
        
        # attachment_names
        if self.attachment_names != other.attachment_names:
            return False
        
        # author_ids
        if self.author_ids != other.author_ids:
            return False
        
        # author_types
        if self.author_types != other.author_types:
            return False
        
        # before
        if self.before != other.before:
            return False
        
        # channel_ids
        if self.channel_ids != other.channel_ids:
            return False
        
        # content
        if self.content != other.content:
            return False
        
        # embed_providers
        if self.embed_providers != other.embed_providers:
            return False
        
        # embed_types
        if self.embed_types != other.embed_types:
            return False
        
        # has_types
        if self.has_types != other.has_types:
            return False
        
        # include_nsfw_channels
        if self.include_nsfw_channels != other.include_nsfw_channels:
            return False
        
        # limit
        if self.limit != other.limit:
            return False
        
        # mentioned_everyone
        if self.mentioned_everyone != other.mentioned_everyone:
            return False
        
        # mentioned_user_ids
        if self.mentioned_user_ids != other.mentioned_user_ids:
            return False
        
        # offset
        if self.offset != other.offset:
            return False
        
        # pinned
        if self.pinned != other.pinned:
            return False
        
        # slop
        if self.slop != other.slop:
            return False
        
        # sort_by
        if self.sort_by is not other.sort_by:
            return False
        
        # sort_order
        if self.sort_order is not other.sort_order:
            return False
        
        # url_host_names
        if self.url_host_names != other.url_host_names:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the message search query.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new._serialisation_flags = self._serialisation_flags
        
        # after
        new.after = self.after
        
        # attachment_extensions
        attachment_extensions = self.attachment_extensions
        if (attachment_extensions is not None):
            attachment_extensions = (*attachment_extensions,)
        new.attachment_extensions = attachment_extensions
        
        # attachment_names
        attachment_names = self.attachment_names
        if (attachment_names is not None):
            attachment_names = (*attachment_names,)
        new.attachment_names = attachment_names
        
        # author_ids
        author_ids = self.author_ids
        if (author_ids is not None):
            author_ids = (*author_ids,)
        new.author_ids = author_ids
        
        # author_types
        author_types = self.author_types
        if (author_types is not None):
            author_types = (*author_types,)
        new.author_types = author_types
        
        # before
        new.before = self.before
        
        # channel_ids
        channel_ids = self.channel_ids
        if (channel_ids is not None):
            channel_ids = (*channel_ids,)
        new.channel_ids = channel_ids
        
        # content
        new.content = self.content
        
        # embed_providers
        embed_providers = self.embed_providers
        if (embed_providers is not None):
            embed_providers = (*embed_providers,)
        new.embed_providers = embed_providers
        
        # embed_types
        embed_types = self.embed_types
        if (embed_types is not None):
            embed_types = (*embed_types,)
        new.embed_types = embed_types
        
        # has_types
        has_types = self.has_types
        if (has_types is not None):
            has_types = (*has_types,)
        new.has_types = has_types
        
        # include_nsfw_channels
        new.include_nsfw_channels = self.include_nsfw_channels
        
        # limit
        new.limit = self.limit
        
        # mentioned_everyone
        new.mentioned_everyone = self.mentioned_everyone
        
        # mentioned_user_ids
        mentioned_user_ids = self.mentioned_user_ids
        if (mentioned_user_ids is not None):
            mentioned_user_ids = (*mentioned_user_ids,)
        new.mentioned_user_ids = mentioned_user_ids
        
        # offset
        new.offset = self.offset
        
        # pinned
        new.pinned = self.pinned
        
        # slop
        new.slop = self.slop
        
        # sort_by
        new.sort_by = self.sort_by
        
        # sort_order
        new.sort_order = self.sort_order
        
        # url_host_names
        url_host_names = self.url_host_names
        if (url_host_names is not None):
            url_host_names = (*url_host_names,)
        new.url_host_names = url_host_names
        
        return new
    
    
    def copy_with(
        self,
        *,
        after = ...,
        attachment_extensions = ...,
        attachment_names = ...,
        author_ids = ...,
        author_types = ...,
        before = ...,
        channel_ids = ...,
        content = ...,
        embed_providers = ...,
        embed_types = ...,
        has_types = ...,
        include_nsfw_channels = ...,
        limit = ...,
        mentioned_everyone = ...,
        mentioned_user_ids = ...,
        offset = ...,
        pinned = ...,
        slop = ...,
        sort_by = ...,
        sort_order = ...,
        url_host_names = ...,
    ):
        """
        Copies the message search query with the given fields.
        
        Parameters
        ----------
        after : `None | DateTime`, Optional (Keyword only)
            Message creation time lower threshold.
        
        attachment_extensions : `None | str | iterable<str>`, Optional (Keyword only)
            Whether the returned messages should have attachments with these extensions.
        
        attachment_names : `None | str | iterable<str>`, Optional (Keyword only)
            Whether the returned messages should have attachments with these names.
        
        author_ids : ``None | int | iterable<int> | UserBase | iterable<UserBase>``, Optional (Keyword only)
            Whether the returned messages should be created by these users defined by their identifiers.
        
        author_types : ``None | str | iterable<str> | MessageSearchAuthorType | iterable<MessageSearchAuthorType>`` \
                , Optional (Keyword only)
            Whether the returned messages should be created by these type of users.
        
        before : `None | DateTime`, Optional (Keyword only)
            Message creation time upper threshold.
        
        channel_ids : ``None | int | iterable<int> | Channel | iterable<Channel>``, Optional (Keyword only)
            Whether the returned messages should be created in these channels defined by their identifiers.
        
        content : `None | str`, Optional (Keyword only)
            Content to search in the messages.
        
        embed_providers : `None | str | iterable<str>`, Optional (Keyword only)
            Whether the returned messages should have embeds with these providers.
        
        embed_types : ``None | str | iterable<str> | EmbedType | iterable<EmbedType>``, Optional (Keyword only)
            Whether the returned messages should have embeds with these types.
        
        has_types : ``None | str | iterable<str> | MessageSearchHasType | iterable<MessageSearchHasType>`` \
                , Optional (Keyword only)
            What type of content should the returned messages have.
        
        include_nsfw_channels : `None | bool`, Optional (Keyword only)
            Whether to include nsfw channels in the search.
        
        limit : `None | int`, Optional (Keyword only)
            The amount of messages to return.
        
        mentioned_everyone : `None | bool`, Optional (Keyword only)
            Whether the returned messages should be mentioning everyone.
        
        mentioned_user_ids : ``None | int | iterable<int> | ClientUserBase | iterable<ClientUserBase>`` \
                , Optional (Keyword only)
            Whether the returned messages should be mentioned these users by their identifiers.
        
        offset : `None | int`, Optional (Keyword only)
            Amount to offset the returned messages by.
        
        pinned : `None | bool`, Optional (Keyword only)
            Whether the returned messages should be pinned.
        
        slop : `None | int`, Optional (Keyword only)
            The maximal amount of words that can be in between each word from `content`.
        
        sort_by : ``None | str | MessageSearchSortByType``, Optional (Keyword only)
            By what to sort the returned messages.
        
        sort_order : ``None | str | MessageSearchSortOrderType``, Optional (Keyword only)
            How to order the returned messages. Depending on the `sort_by`, this field may not be respected.
        
        url_host_names : `None | str | iterable<str>`, Optional (Keyword only)
            Whether the returned messages should contain urls with these host names.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        ValueError
            - If a parameter's value is incorrect.
        TypeError
            - If a parameter's type is incorrect.
        """
        serialisation_flags = self._serialisation_flags
        
        # after
        if (after is ...):
            after = self.after
        else:
            after = validate_after(after)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_AFTER)
        
        # attachment_extensions
        if (attachment_extensions is ...):
            attachment_extensions = self.attachment_extensions
            if (attachment_extensions is not None):
                attachment_extensions = (*attachment_extensions,)
        else:
            attachment_extensions = validate_attachment_extensions(attachment_extensions)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_ATTACHMENT_EXTENSIONS)
        
        # attachment_names
        if (attachment_names is ...):
            attachment_names = self.attachment_names
            if (attachment_names is not None):
                attachment_names = (*attachment_names,)
        else:
            attachment_names = validate_attachment_names(attachment_names)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_ATTACHMENT_NAMES)
        
        # author_ids
        if (author_ids is ...):
            author_ids = self.author_ids
            if (author_ids is not None):
                author_ids = (*author_ids,)
        else:
            author_ids = validate_author_ids(author_ids)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_AUTHOR_IDS)
        
        # author_types
        if (author_types is ...):
            author_types = self.author_types
            if (author_types is not None):
                author_types = (*author_types,)
        else:
            author_types = validate_author_types(author_types)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_AUTHOR_TYPES)
        
        # before
        if (before is ...):
            before = self.before
        else:
            before = validate_before(before)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_BEFORE)
        
        # channel_ids
        if (channel_ids is ...):
            channel_ids = self.channel_ids
            if (channel_ids is not None):
                channel_ids = (*channel_ids,)
        else:
            channel_ids = validate_channel_ids(channel_ids)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_CHANNEL_IDS)
        
        # content
        if (content is ...):
            content = self.content
        else:
            content = validate_content(content)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_CONTENT)
        
        # embed_providers
        if (embed_providers is ...):
            embed_providers = self.embed_providers
            if (embed_providers is not None):
                embed_providers = (*embed_providers,)
        else:
            embed_providers = validate_embed_providers(embed_providers)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_EMBED_PROVIDERS)
        
        # embed_types
        if (embed_types is ...):
            embed_types = self.embed_types
            if (embed_types is not None):
                embed_types = (*embed_types,)
        else:
            embed_types = validate_embed_types(embed_types)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_EMBED_TYPES)
        
        # has_types
        if (has_types is ...):
            has_types = self.has_types
            if (has_types is not None):
                has_types = (*has_types,)
        else:
            has_types = validate_has_types(has_types)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_HAS_TYPES)
        
        # include_nsfw_channels
        if (include_nsfw_channels is ...):
            include_nsfw_channels = self.include_nsfw_channels
        else:
            include_nsfw_channels = validate_include_nsfw_channels(include_nsfw_channels)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_INCLUDE_NSFW_CHANNELS)
        
        # limit
        if (limit is ...):
            limit = self.limit
        else:
            limit = validate_limit(limit)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_LIMIT)
        
        # mentioned_everyone
        if (mentioned_everyone is ...):
            mentioned_everyone = self.mentioned_everyone
        else:
            mentioned_everyone = validate_mentioned_everyone(mentioned_everyone)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_MENTIONED_EVERYONE)
        
        # mentioned_user_ids
        if (mentioned_user_ids is ...):
            mentioned_user_ids = self.mentioned_user_ids
            if (mentioned_user_ids is not None):
                mentioned_user_ids = (*mentioned_user_ids,)
        else:
            mentioned_user_ids = validate_mentioned_user_ids(mentioned_user_ids)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_MENTIONED_USER_IDS)
        
        # offset
        if (offset is ...):
            offset = self.offset
        else:
            offset = validate_offset(offset)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_OFFSET)
        
        # pinned
        if (pinned is ...):
            pinned = self.pinned
        else:
            pinned = validate_pinned(pinned)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_PINNED)
        
        # slop
        if (slop is ...):
            slop = self.slop
        else:
            slop = validate_slop(slop)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_SLOP)
        
        # sort_by
        if (sort_by is ...):
            sort_by = self.sort_by
        else:
            sort_by = validate_sort_by(sort_by)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_SORT_BY)
        
        # sort_order
        if (sort_order is ...):
            sort_order = self.sort_order
        else:
            sort_order = validate_sort_order(sort_order)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_SORT_ORDER)
        
        # url_host_names
        if (url_host_names is ...):
            url_host_names = self.url_host_names
            if (url_host_names is not None):
                url_host_names = (*url_host_names,)
        else:
            url_host_names = validate_url_host_names(url_host_names)
            serialisation_flags |= (1 << SERIALISATION_SHIFT_URL_HOST_NAMES)
        
        # Construct
        new = object.__new__(type(self))
        new._serialisation_flags = serialisation_flags
        new.after = after
        new.attachment_extensions = attachment_extensions
        new.attachment_names = attachment_names
        new.author_ids = author_ids
        new.author_types = author_types
        new.before = before
        new.channel_ids = channel_ids
        new.content = content
        new.embed_providers = embed_providers
        new.embed_types = embed_types
        new.has_types = has_types
        new.include_nsfw_channels = include_nsfw_channels
        new.limit = limit
        new.mentioned_everyone = mentioned_everyone
        new.mentioned_user_ids = mentioned_user_ids
        new.offset = offset
        new.pinned = pinned
        new.slop = slop
        new.sort_by = sort_by
        new.sort_order = sort_order
        new.url_host_names = url_host_names
        return new
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a message search query from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data to create the instance from.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._serialisation_flags = (
            (('min_id' in data) << SERIALISATION_SHIFT_AFTER) |
            (('attachment_extension' in data) << SERIALISATION_SHIFT_ATTACHMENT_EXTENSIONS) |
            (('attachment_filename' in data) << SERIALISATION_SHIFT_ATTACHMENT_NAMES) |
            (('author_id' in data) << SERIALISATION_SHIFT_AUTHOR_IDS) |
            (('author_type' in data) << SERIALISATION_SHIFT_AUTHOR_TYPES) |
            (('max_id' in data) << SERIALISATION_SHIFT_BEFORE) |
            (('channel_id' in data) << SERIALISATION_SHIFT_CHANNEL_IDS) |
            (('content' in data) << SERIALISATION_SHIFT_CONTENT) |
            (('embed_provider' in data) << SERIALISATION_SHIFT_EMBED_PROVIDERS) |
            (('embed_type' in data) << SERIALISATION_SHIFT_EMBED_TYPES) |
            (('has' in data) << SERIALISATION_SHIFT_HAS_TYPES) |
            (('include_nsfw' in data) << SERIALISATION_SHIFT_INCLUDE_NSFW_CHANNELS) |
            (('limit' in data) << SERIALISATION_SHIFT_LIMIT) |
            (('mention_everyone' in data) << SERIALISATION_SHIFT_MENTIONED_EVERYONE) |
            (('mentions' in data) << SERIALISATION_SHIFT_MENTIONED_USER_IDS) |
            (('offset' in data) << SERIALISATION_SHIFT_OFFSET) |
            (('pinned' in data) << SERIALISATION_SHIFT_PINNED) |
            (('slop' in data) << SERIALISATION_SHIFT_SLOP) |
            (('sort_by' in data) << SERIALISATION_SHIFT_SORT_BY) |
            (('sort_order' in data) << SERIALISATION_SHIFT_SORT_ORDER) |
            (('link_hostname' in data) << SERIALISATION_SHIFT_URL_HOST_NAMES)
        )
        self.after = parse_after(data)
        self.attachment_extensions = parse_attachment_extensions(data)
        self.attachment_names = parse_attachment_names(data)
        self.author_ids = parse_author_ids(data)
        self.author_types = parse_author_types(data)
        self.before = parse_before(data)
        self.channel_ids = parse_channel_ids(data)
        self.content = parse_content(data)
        self.embed_providers = parse_embed_providers(data)
        self.embed_types = parse_embed_types(data)
        self.has_types = parse_has_types(data)
        self.include_nsfw_channels = parse_include_nsfw_channels(data)
        self.limit = parse_limit(data)
        self.mentioned_everyone = parse_mentioned_everyone(data)
        self.mentioned_user_ids = parse_mentioned_user_ids(data)
        self.offset = parse_offset(data)
        self.pinned = parse_pinned(data)
        self.slop = parse_slop(data)
        self.sort_by = parse_sort_by(data)
        self.sort_order = parse_sort_order(data)
        self.url_host_names = parse_url_host_names(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serialises the message search query.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields as their defaults should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        serialisation_flags = self._serialisation_flags
        put_after(
            self.after,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_AFTER) & 1),
        )
        put_attachment_extensions(
            self.attachment_extensions,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_ATTACHMENT_EXTENSIONS) & 1),
        )
        put_attachment_names(
            self.attachment_names,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_ATTACHMENT_NAMES) & 1),
        )
        put_author_ids(
            self.author_ids,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_AUTHOR_IDS) & 1),
        )
        put_author_types(
            self.author_types,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_AUTHOR_TYPES) & 1),
        )
        put_before(
            self.before,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_BEFORE) & 1),
        )
        put_channel_ids(
            self.channel_ids,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_CHANNEL_IDS) & 1),
        )
        put_content(
            self.content,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_CONTENT) & 1),
        )
        put_embed_providers(
            self.embed_providers,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_EMBED_PROVIDERS) & 1),
        )
        put_embed_types(
            self.embed_types,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_EMBED_TYPES) & 1),
        )
        put_has_types(
            self.has_types,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_HAS_TYPES) & 1),
        )
        put_include_nsfw_channels(
            self.include_nsfw_channels,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_INCLUDE_NSFW_CHANNELS) & 1),
        )
        put_limit(
            self.limit,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_LIMIT) & 1),
        )
        put_mentioned_everyone(
            self.mentioned_everyone,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_MENTIONED_EVERYONE) & 1),
        )
        put_mentioned_user_ids(
            self.mentioned_user_ids,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_MENTIONED_USER_IDS) & 1),
        )
        put_offset(
            self.offset,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_OFFSET) & 1),
        )
        put_pinned(
            self.pinned,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_PINNED) & 1),
        )
        put_slop(
            self.slop,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_SLOP) & 1),
        )
        put_sort_by(
            self.sort_by,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_SORT_BY) & 1),
        )
        put_sort_order(
            self.sort_order,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_SORT_ORDER) & 1),
        )
        put_url_host_names(
            self.url_host_names,
            data,
            defaults or ((serialisation_flags >> SERIALISATION_SHIFT_URL_HOST_NAMES) & 1),
        )
        return data



__all__ = ('BanAddMultipleResult',)

from scarletio import RichAttributeErrorBaseType

from ...user import create_partial_user_from_id

from .fields import (
    parse_banned_user_ids, parse_failed_user_ids, put_banned_user_ids, put_failed_user_ids,
    validate_banned_user_ids, validate_failed_user_ids
)


class BanAddMultipleResult(RichAttributeErrorBaseType):
    """
    A ban add multiple result.
    
    Attributes
    ----------
    banned_user_ids : `None | tuple<int>`
        The actually banned users' identifiers.
    failed_user_ids : `None | tuple<int>`
        The failed to ban users' identifiers.
    """
    __slots__ = ('banned_user_ids', 'failed_user_ids',)
    
    def __new__(cls, *, banned_user_ids = ..., failed_user_ids = ...):
        """
        Creates a new ban add multiple result instance.
        
        Parameters
        ----------
        banned_user_ids : `None | iterable<int | ClientUserBase>`, Optional (Keyword only)
            The actually banned users' identifiers.
        failed_user_ids : `None | iterable<int | ClientUserBase>`, Optional (Keyword only)
            The failed to ban users' identifiers.
        
        Raises
        -------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # banned_user_ids
        if banned_user_ids is ...:
            banned_user_ids = None
        else:
            banned_user_ids = validate_banned_user_ids(banned_user_ids)
        
        # failed_user_ids
        if failed_user_ids is ...:
            failed_user_ids = None
        else:
            failed_user_ids = validate_failed_user_ids(failed_user_ids)
        
        # Construct
        self = object.__new__(cls)
        self.banned_user_ids = banned_user_ids
        self.failed_user_ids = failed_user_ids
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a ban add multiple result instance from the given `data`.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data representing a ban add multiple result.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.banned_user_ids = parse_banned_user_ids(data)
        self.failed_user_ids = parse_failed_user_ids(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the ban add multiple result into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_banned_user_ids(self.banned_user_ids, data, defaults)
        put_failed_user_ids(self.failed_user_ids, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the ban add multiple result's representation."""
        repr_parts = ['<', type(self).__name__]
        
        
        # banned_user_ids
        banned_user_ids = self.banned_user_ids
        if (banned_user_ids is not None):
            repr_parts.append(' banned_user_ids = ')
            repr_parts.append(repr(self.banned_user_ids))
            
            
            field_added = True
        else:
            field_added = False
        
        # failed_user_ids
        failed_user_ids = self.failed_user_ids
        if (failed_user_ids is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' failed_user_ids = ')
            repr_parts.append(repr(failed_user_ids))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __len__(self):
        """Helper for unpacking."""
        return 2
    
    
    def __iter__(self):
        """Unpacks the ban add multiple result."""
        yield self.banned_user_ids
        yield self.failed_user_ids

    
    def __hash__(self):
        """Returns the ban add multiple result's hash value."""
        hash_value = 0
        
        # banned_user_ids
        banned_user_ids = self.banned_user_ids
        if (banned_user_ids is not None):
            hash_value ^= len(banned_user_ids)
            for user_id in banned_user_ids:
                hash_value ^= user_id
        
        # failed_user_ids
        failed_user_ids = self.failed_user_ids
        if (failed_user_ids is not None):
            hash_value ^= len(failed_user_ids) << 8
            for user_id in failed_user_ids:
                hash_value ^= user_id
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two ban entries are equal."""
        if type(self) is not type(other):
            return False
        
        # reason
        if self.banned_user_ids != other.banned_user_ids:
            return False
        
        # user
        if self.failed_user_ids != other.failed_user_ids:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the ban add multiple result.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        banned_user_ids = self.banned_user_ids
        if (banned_user_ids is not None):
            banned_user_ids = (*banned_user_ids,)
        new.banned_user_ids = banned_user_ids
        
        failed_user_ids = self.failed_user_ids
        if (failed_user_ids is not None):
            failed_user_ids = (*failed_user_ids,)
        new.failed_user_ids = failed_user_ids
        
        return new
    
    
    def copy_with(self, *, banned_user_ids = ..., failed_user_ids = ...):
        """
        Copies the ban add multiple result.
        
        Parameters
        ----------
        banned_user_ids : `None | iterable<int | ClientUserBase>`, Optional (Keyword only)
            The actually banned users' identifiers.
        failed_user_ids : `None | iterable<int | ClientUserBase>`, Optional (Keyword only)
            The failed to ban users' identifiers.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        -------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # banned_user_ids
        if banned_user_ids is ...:
            banned_user_ids = self.banned_user_ids
            if (banned_user_ids is not None):
                banned_user_ids = (*banned_user_ids,)
        else:
            banned_user_ids = validate_banned_user_ids(banned_user_ids)
        
        # failed_user_ids
        if failed_user_ids is ...:
            failed_user_ids = self.failed_user_ids
            if (failed_user_ids is not None):
                failed_user_ids = (*failed_user_ids,)
        else:
            failed_user_ids = validate_failed_user_ids(failed_user_ids)
        
        # Construct
        new = object.__new__(type(self))
        new.banned_user_ids = banned_user_ids
        new.failed_user_ids = failed_user_ids
        return new
    
    
    def iter_banned_user_ids(self):
        """
        Iterates over the banned users' identifiers.
        
        This method is an iterable generator.
        
        yields
        ------
        user_id : `int`
        """
        banned_user_ids = self.banned_user_ids
        if (banned_user_ids is not None):
            yield from banned_user_ids
    
    
    def iter_banned_users(self):
        """
        Iterates over the banned users.
        
        This method is an iterable generator.
        
        yields
        ------
        user : ``ClientUserBase`
        """
        banned_user_ids = self.banned_user_ids
        if (banned_user_ids is not None):
            for user_id in banned_user_ids:
                yield create_partial_user_from_id(user_id)
    
    
    def iter_failed_user_ids(self):
        """
        Iterates over the failed to ban users' identifiers.
        
        This method is an iterable generator.
        
        yields
        ------
        user_id : `int`
        """
        failed_user_ids = self.failed_user_ids
        if (failed_user_ids is not None):
            yield from failed_user_ids
    
    
    def iter_failed_users(self):
        """
        Iterates over the failed to ban users.
        
        This method is an iterable generator.
        
        yields
        ------
        user : ``ClientUserBase`
        """
        failed_user_ids = self.failed_user_ids
        if (failed_user_ids is not None):
            for user_id in failed_user_ids:
                yield create_partial_user_from_id(user_id)



__all__ = ()

from ...field_parsers import functional_parser_factory
from ...field_putters import functional_putter_factory
from ...field_validators import entity_validator_factory, preinstanced_validator_factory

from ..emoji import Emoji, create_partial_emoji_data, create_partial_emoji_from_data

from .preinstanced import ReactionType


# emoji

parse_emoji = functional_parser_factory('emoji', create_partial_emoji_from_data)
put_emoji = functional_putter_factory('emoji', create_partial_emoji_data)
validate_emoji = entity_validator_factory('emoji', Emoji)

# type

def parse_type(data):
    """
    Parses out the reaction's type the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Reaction data.
    
    Returns
    -------
    reaction_type : ``ReactionType``
    """
    if data.get('burst', False):
        return ReactionType.burst
    
    return ReactionType.standard


def put_type(reaction_type, data, defaults):
    """
    Puts the reaction's type into the given `data` json serializable object.
    
    Parameters
    ----------
    reaction_type : ``ReactionType``
        The reaction's type.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if reaction_type is ReactionType.burst:
        burst_field_value = True
    else:
        burst_field_value = False
        
    data['burst'] = burst_field_value
    return data


validate_type = preinstanced_validator_factory('reaction_type', ReactionType)



__all__ = ('AuditLogEntry', )

from ...bases import DiscordEntity
from ...core import GUILDS
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...user import create_partial_user_from_id

from ..conversion_helpers.helpers import _hash_dict

from .fields import (
    parse_changes, parse_details, parse_guild_id, parse_id, parse_reason, parse_target_id, parse_type, parse_user_id,
    put_changes, put_details, put_guild_id, put_id, put_reason, put_target_id,
    put_type, put_user_id, validate_changes, validate_details, validate_guild_id, validate_id,
    validate_parent, validate_reason, validate_target_id, validate_type, validate_user_id
)
from .preinstanced import AuditLogEntryType


PRECREATE_FIELDS = {
    'guild': ('guild_id', validate_guild_id),
    'guild_id': ('guild_id', validate_guild_id),
    'reason': ('reason', validate_reason),
    'user': ('user_id', validate_user_id),
    'user_id': ('user_id', validate_user_id),
    'target_id': ('target_id', validate_target_id),
}


class AuditLogEntry(DiscordEntity):
    """
    Represents an entry of an audit log.
    
    Attributes
    ----------
    _parent_reference : `None`, ``WeakReferer`` to ``AuditLog``
        Reference to the audit log entry keys' parent.
    
    changes : `None | dict<str, AuditLogChange>`
        The changes of the entry.
    
    details : `None | dict<str, object>`
        Additional information for the specific action types.
    
    guild_id : `int`
        The source guild's identifier.
    
    id : `int`
        The unique identifier number of the entry.
    
    reason : `None`, `str`
        The reason provided with the logged action.
    
    target_id : `int`
        The entry's target's identifier.
        
        If no target is provided, this value defaults to `0`.
    
    type : ``AuditLogEntryType``
        The logged event's type.
    
    user_id : `int`
        The user's identifier who triggered the event.
    """
    __slots__ = ('_parent_reference', 'changes', 'details', 'guild_id', 'reason', 'target_id', 'type', 'user_id',)
    
    
    def __new__(
        cls,
        *,
        changes = ...,
        details = ...,
        entry_type = ...,
        guild_id = ...,
        reason = ...,
        target_id = ...,
        user_id = ...,
    ):
        """
        Creates a new audit log entry with the given fields.
        
        Parameters
        ----------
        changes : `None | iterable<AuditLogChange>`, Optional (Keyword only)
            The changes of the entry.
        
        details : `None | dict<str, object>`, Optional (Keyword only)
            Additional information for the specific action types.
        
        entry_type : ``AuditLogEntryType``, `int`, Optional (Keyword only)
            The logged event's type.
        
        guild_id : `int`, `None`, ``Guild``, Optional (Keyword only)
            The source guild's identifier.
        
        reason : `None`, `str`, Optional (Keyword only)
            The reason provided with the logged action.
        
        target_id : `int`, Optional (Keyword only)
            The entry's target's identifier.
        
        user_id : `int`, ``None | ClientUserBase```, Optional (Keyword only)
            The user's identifier who triggered the event.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # entry_type
        if entry_type is ...:
            entry_type = AuditLogEntryType.none
        else:
            entry_type = validate_type(entry_type)
        
        # changes
        if changes is ...:
            changes = None
        else:
            changes = validate_changes(changes, entry_type = entry_type)
        
        # details
        if details is ...:
            details = None
        else:
            details = validate_details(details, entry_type = entry_type)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # reason
        if reason is ...:
            reason = None
        else:
            reason = validate_reason(reason)
        
        # target_id
        if target_id is ...:
            target_id = 0
        else:
            target_id = validate_target_id(target_id)
        
        # user_id
        if user_id is ...:
            user_id = 0
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        self = object.__new__(cls)
        self._parent_reference = None
        self.changes = changes
        self.details = details
        self.guild_id = guild_id
        self.id = 0
        self.reason = reason
        self.target_id = target_id
        self.type = entry_type
        self.user_id = user_id
        return self
    
        
    @classmethod
    def from_data(cls, data, parent = None):
        """
        Creates an audit log entry, from entry data sent inside of an ``AuditLog``'s data and from the audit itself.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data received from Discord.
        
        parent : `None`, ``AuditLog`` = `None`, Optional
            The parent of the entry that contains the respective guild, the included users, webhooks and the
            integrations, etc... to work with.
        
        Returns
        -------
        self : `instance<cls>`
            Returns `None` if Discord is derping like hell.
        """
        # type
        entry_type = parse_type(data)
        if entry_type is AuditLogEntryType.none:
            return
        
        # guild_id
        guild_id = parse_guild_id(data, parent)
        
        # parent reference
        if parent is None:
            parent_reference = None
        else:
            parent_reference = parent._get_self_reference()
        
        # Construct
        self = object.__new__(cls)
        self._parent_reference = parent_reference
        self.changes = parse_changes(data, entry_type)
        self.details = parse_details(data, entry_type)
        self.guild_id = guild_id
        self.id = parse_id(data)
        self.reason = parse_reason(data)
        self.target_id = parse_target_id(data)
        self.type = entry_type
        self.user_id = parse_user_id(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serialises the audit log entry.
        
        Parameters
        ----------
        defaults : `bool`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        entry_type = self.type
        put_changes(self.changes, data, defaults, entry_type = entry_type)
        put_details(self.details, data, defaults, entry_type = entry_type)
        put_guild_id(self.guild_id, data, defaults)
        put_id(self.id, data, defaults)
        put_reason(self.reason, data, defaults)
        put_target_id(self.target_id, data, defaults)
        put_type(entry_type, data, defaults)
        put_user_id(self.user_id, data, defaults)
        return data
    
    
    @classmethod
    def precreate(cls, entry_id, *, changes = ..., details = ..., entry_type = ..., parent = ..., **keyword_parameters):
        """
        Creates a new audit log entry with the given fields.
        
        Parameters
        ----------
        entry_id : `int`
            The entry's unique identifier.
        
        changes : `None | iterable<AuditLogChange>`, Optional (Keyword only)
            The changes of the entry.
        
        details : `None | dict<str, object>`, Optional (Keyword only)
            Additional information for the specific action types.
        
        entry_type : ``AuditLogEntryType``, `int`, Optional (Keyword only)
            The logged event's type.
        
        parent : `None | AuditLog`, Optional (Keyword only)
            The parent audit log for the entry.
        
        **keyword_parameters : Keyword Parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        
        guild : `int`, ``None | Guild``, Optional (Keyword only)
            Alternative for `guild_id`.
        
        guild_id : `int`, `None`, ``Guild``, Optional (Keyword only)
            The source guild's identifier.
        
        reason : `None`, `str`, Optional (Keyword only)
            The reason provided with the logged action.
        
        target_id : `int`, Optional (Keyword only)
            The entry's target's identifier.
        
        user : `int`, ``None | ClientUserBase```, Optional (Keyword only)
            Alternative for `user_id`.
        
        user_id : `int`, ``None | ClientUserBase```, Optional (Keyword only)
            The user's identifier who triggered the event.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        entry_id = validate_id(entry_id)
        
        # entry_type
        if entry_type is ...:
            entry_type = AuditLogEntryType.none
        else:
            entry_type = validate_type(entry_type)
        
        # changes
        if changes is ...:
            changes = None
        else:
            changes = validate_changes(changes, entry_type = entry_type)
        
        # details
        if details is ...:
            details = None
        else:
            details = validate_details(details, entry_type = entry_type)
        
        # parent
        if parent is ...:
            parent = None
        else:
            parent = validate_parent(parent)
        
        # keyword_parameters
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        # Get parent reference
        if parent is None:
            parent_reference = None
        else:
            parent_reference = parent._get_self_reference()
        
        # Construct
        self = object.__new__(cls)
        self._parent_reference = parent_reference
        self.changes = changes
        self.details = details
        self.guild_id = 0
        self.id = entry_id
        self.reason = None
        self.target_id = 0
        self.type = entry_type
        self.user_id = 0
        
        # Set processed fields
        if (processed is not None):
            for name, value in processed:
                setattr(self, name, value)
        
        return self
    
    
    def copy(self):
        """
        Copies the audit log entry.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new._parent_reference = None
        
        changes = self.changes
        if (changes is not None):
            changes = changes.copy()
        new.changes = changes
        
        details = self.details
        if (details is not None):
            details = details.copy()
        new.details = details
        
        new.guild_id = self.guild_id
        new.id = 0
        new.reason = self.reason
        new.target_id = self.target_id
        new.type = self.type
        new.user_id = self.user_id
        return new
    
    
    def copy_with(
        self, 
        *,
        changes = ...,
        details = ...,
        entry_type = ...,
        guild_id = ...,
        reason = ...,
        target_id = ...,
        user_id = ...,
    ):
        """
        Copies the audit log entry with the given fields.
        
        Parameters
        ----------
        changes : `None | iterable<AuditLogChange>`, Optional (Keyword only)
            The changes of the entry.
        
        details : `None | dict<str, object>`, Optional (Keyword only)
            Additional information for the specific action types.
        
        entry_type : ``AuditLogEntryType``, `int`, Optional (Keyword only)
            The logged event's type.
        
        guild_id : `int`, `None`, ``Guild``, Optional (Keyword only)
            The source guild's identifier.
        
        reason : `None`, `str`, Optional (Keyword only)
            The reason provided with the logged action.
        
        target_id : `int`, Optional (Keyword only)
            The entry's target's identifier.
        
        user_id : `int`, ``None | ClientUserBase```, Optional (Keyword only)
            The user's identifier who triggered the event.
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # We yeet old changes and details on `target_type` mismatch.
        # entry_type
        if entry_type is ...:
            entry_type = self.type
            target_type_mismatch = False
        else:
            entry_type = validate_type(entry_type)
            target_type_mismatch = self.type.target_type is not entry_type.target_type
        
        # changes
        if changes is ...:
            if target_type_mismatch:
                changes = None
            else:
                changes = self.changes
                if (changes is not None):
                    changes = changes.copy()
        else:
            changes = validate_changes(changes, entry_type = entry_type)
        
        # details
        if details is ...:
            if target_type_mismatch:
                details = None
            else:
                details = self.details
                if (details is not None):
                    details = details.copy()
        else:
            details = validate_details(details, entry_type = entry_type)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # reason
        if reason is ...:
            reason = self.reason
        else:
            reason = validate_reason(reason)
        
        # target_id
        if target_id is ...:
            target_id = self.target_id
        else:
            target_id = validate_target_id(target_id)
        
        # user_id
        if user_id is ...:
            user_id = self.user_id
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        self = object.__new__(type(self))
        self._parent_reference = None
        self.changes = changes
        self.details = details
        self.guild_id = guild_id
        self.id = 0
        self.reason = reason
        self.target_id = target_id
        self.type = entry_type
        self.user_id = user_id
        return self
    
    
    @property
    def target(self):
        """
        Tries to resolve the target entity of the audit log entry.
        
        If the entity is not cached, returns `None`.
        
        Returns
        -------
        target : `None`, ``Guild``,  ``Channel``, ``ClientUserBase``, ``Role``, ``Webhook``,
            ``Emoji``, ``Message``, ``Integration``, ``Sticker``, ``Stage``, ``ScheduledEvent``
        """
        target_converter = self.type.target_type.target_converter
        if (target_converter is not None):
            return target_converter(self)
    
    
    @property
    def parent(self):
        """
        Returns the entry's parent audit log.
        
        Returns
        -------
        parent : `None`, ``AuditLog``
        """
        parent_reference = self._parent_reference
        if (parent_reference is not None):
            return parent_reference()
    
    
    @property
    def guild(self):
        """
        Returns the audit log's guild.
        
        Returns
        -------
        guild : ``None | Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def user(self):
        """
        Returns the user who triggered the event's creation.
        
        Returns
        -------
        user : ``None | ClientUserBase``
            Returns ˙None` if not an user triggered it.
        """
        user_id = self.user_id
        if user_id:
            return create_partial_user_from_id(user_id)
    
    
    def __repr__(self):
        """Returns the representation of the audit log entry."""
        repr_parts = ['<', self.__class__.__name__]
        
        # id
        entry_id = self.id
        if entry_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(entry_id))
            repr_parts.append(',')
        
        # type
        entry_type = self.type
        repr_parts.append(' type = ')
        repr_parts.append(entry_type.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(entry_type.value))
        
        # user_id
        user_id = self.user_id
        if user_id:
            repr_parts.append(', user_id = ')
            repr_parts.append(repr(user_id))
        
        # guild_id
        guild_id = self.guild_id
        if guild_id:
            repr_parts.append(', guild_id = ')
            repr_parts.append(repr(guild_id))
        
        # target_id
        target_id = self.target_id
        if target_id:
            repr_parts.append(', target_id = ')
            repr_parts.append(repr(target_id))
        
        # changes
        changes = self.changes
        if (changes is not None):
            repr_parts.append(', changes = ')
            repr_parts.append(repr(changes))
        
        # details
        details = self.details
        if (details is not None):
            repr_parts.append(', details = ')
            repr_parts.append(repr(details))
        
        # reason
        reason = self.reason
        if reason is not None:
            repr_parts.append(', reason = ')
            repr_parts.append(repr(reason))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two audit log entries are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two audit log entries are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two audit log entries are equal.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance. Must be the same type as self.
        
        Returns
        -------
        is_equal : `bool`
        """
        # id
        self_id = self.id
        other_id = other.id
        if self_id and other_id and self_id != other_id:
            return False
        
        # changes
        if self.changes != other.changes:
            return False
        
        # details
        if self.details != other.details:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        # reason
        if self.reason != other.reason:
            return False
        
        # target_id
        if self.target_id != other.target_id:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # user_id
        if self.user_id != other.user_id:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the audit log entry's hash value."""
        hash_value = 0
        
        # id
        hash_value ^= hash(self.id)
        
        # changes
        changes = self.changes
        if (changes is not None):
            hash_value ^= _hash_dict(changes)
        
        # details
        details = self.details
        if (details is not None):
            hash_value ^= _hash_dict(details)
        
        # guild_id
        hash_value ^= self.guild_id
        
        # reason
        reason = self.reason
        if (reason is not None):
            hash_value ^= hash(reason)
        
        # target_id
        hash_value ^= self.target_id
        
        # type
        hash_value ^= hash(self.type)
        
        # user_id:
        hash_value ^= self.user_id
        
        return hash_value
    
    
    def get_change(self, attribute_name):
        """
        Gets the audit log entry's change for the given `attribute_name`.
        
        Parameters
        ----------
        attribute_name : `str`
            The change by its attribute name to get.
        
        Returns
        -------
        change : `None`, ``AuditLogChange``
        """
        changes = self.changes
        if (changes is not None):
            return changes.get(attribute_name, None)
    
    
    def iter_changes(self):
        """
        Iterates over the changes of the audit log entry.
        
        This method is an iterable generator.
        
        Yields
        ------
        change : ``AuditLogChange``
        """
        changes = self.changes
        if (changes is not None):
            yield from changes.values()
    
    
    def get_detail(self, name):
        """
        Gets the audit log entry's detail for the given `name`.
        
        Parameters
        ----------
        name : `str`
            The detail's name.
        
        Returns
        -------
        change : `None | object`
        """
        details = self.details
        if (details is not None):
            return details.get(name, None)
    
    
    def iter_details(self):
        """
        Iterates over the details of the audit log entry.
        
        This method is an iterable generator.
        
        Yields
        ------
        detail : `(str, object)`
        """
        details = self.details
        if (details is not None):
            yield from details.items()
    
    
    def _link_parent_soft(self, parent):
        """
        Links the given parent to the entry if self is not linked yet.
        
        Parameters
        ----------
        parent : ``AuditLog`
            The parent to link.
        """
        if self.parent is None:
            self._parent_reference = parent._get_self_reference()
    
    
    def _link_parent_hard(self, parent):
        """
        Links the given parent to the entry even if self is already linked.
        
        Parameters
        ----------
        parent : ``AuditLog`
            The parent to link.
        """
        self._parent_reference = parent._get_self_reference()



__all__ = ('InteractionComponentMetadataContainer',)

from scarletio import copy_docs

from .row import InteractionComponentMetadataRow
from .fields import validate_components__container


class InteractionComponentMetadataContainer(InteractionComponentMetadataRow):
    """
    Interaction component metadata representing a container component.
    
    Attributes
    ----------
    components : ``None | tuple<InteractionComponent>``
        Sub-components nested inside.
    """
    __slots__ = ()
    
    @copy_docs(InteractionComponentMetadataRow.__new__)
    def __new__(cls, *, components = ...):
        # components
        if components is ...:
            components = None
        else:
            components = validate_components__container(components)
        
        # Construct.
        self = object.__new__(cls)
        self.components = components
        return self
    
    
    @copy_docs(InteractionComponentMetadataRow.copy_with)
    def copy_with(self, *, components = ...):
        # components
        if components is ...:
            components = self.components
            if (components is not None):
                components = (*(component.copy() for component in components),)
        else:
            components = validate_components__container(components)
        
        # Construct.
        new = object.__new__(type(self))
        new.components = components
        return new



__all__ = ('ForumTag', )

from scarletio import copy_docs

from ...bases import DiscordEntity
from ...core import FORUM_TAGS
from ...precreate_helpers import process_precreate_parameters_and_raise_extra

from .fields import (
    parse_emoji, parse_id, parse_moderated, parse_name, put_emoji, put_id, put_moderated, put_name,
    validate_emoji, validate_id, validate_moderated, validate_name
)


PRECREATE_FIELDS = {
    'emoji': ('emoji', validate_emoji),
    'name': ('name', validate_name),
    'moderated': ('moderated', validate_moderated),
}


class ForumTag(DiscordEntity, immortal = True):
    """
    Forum tags can be applied to a thread of a forum channel.
    
    Attributes
    ----------
    emoji : ``None | Emoji``
        The tag's emoji.
    id : `int`
        The tag's identifier.
    name : `str`
        The tag's name.
    moderated : `bool`
        Whether this tag can only be added or removed by a user with `manage_threads` permission.
    """
    __slots__ = ('emoji', 'name', 'moderated')
    
    def __new__(cls, name = ..., *, emoji = ..., moderated = ...):
        """
        Creates a new forum tag instance.
        
        Parameters
        ----------
        name : `str`
            The tag's name.
        emoji : ``None | Emoji``, Optional (Keyword only)
            The tag's emoji.
        moderated : `bool`, Optional (Keyword only)
            Whether this tag can only be added or removed by a user with `manage_threads` permission.
        """
        # emoji
        if emoji is ...:
            emoji = None
        else:
            emoji = validate_emoji(emoji)
        
        # moderated
        if moderated is ...:
            moderated = False
        else:
            moderated = validate_moderated(moderated)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # Construct
        self = object.__new__(cls)
        self.id = 0
        self.emoji = emoji
        self.name = name
        self.moderated = moderated
        return self
    
    
    @copy_docs(DiscordEntity.__hash__)
    def __hash__(self):
        forum_tag_id = self.id
        if forum_tag_id:
            return forum_tag_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Hashes the fields of the forum tag.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            hash_value ^= emoji.id
        
        # name
        hash_value ^= hash(self.name)
        
        # moderated
        hash_value ^= self.moderated
        
        return hash_value
    
    
    @copy_docs(DiscordEntity.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # id_
        forum_tag_id = self.id
        if forum_tag_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(forum_tag_id))
        else:
            repr_parts.append(' partial')
        
        # name
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        # emoji:
        emoji = self.emoji
        if (emoji is not None):
            repr_parts.append(' emoji = ')
            repr_parts.append(repr(emoji))
        
        # moderated
        if self.moderated:
            repr_parts.append(' moderated = True')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(DiscordEntity.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    @copy_docs(DiscordEntity.__ne__)
    def __ne__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two types are equal.
        
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # id
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self_id == other_id:
                return True
            
            return False
        
        # emoji
        if self.emoji != other.emoji:
            return False
        
        # name:
        if self.name != other.name:
            return False
        
        # moderated
        if self.moderated != other.moderated:
            return False
        
        return True
    
    
    @classmethod
    def _create_empty(cls, forum_tag_id):
        """
        Creates an empty forum tag with their default attributes set.
        
        Parameters
        ----------
        forum_tag_id : `int`
            The forum tag's identifier.
        
        Returns
        -------
        self : ``ForumTag``
        """
        self = object.__new__(cls)
        self.id = forum_tag_id
        self.emoji = None
        self.name = ''
        self.moderated = False
        return self
    
    
    def copy(self):
        """
        Copies the forum tag.
        
        Returns
        -------
        new : ``ForumTag``
        """
        new = object.__new__(type(self))
        
        # id
        new.id = 0
        
        # emoji
        new.emoji = self.emoji
        
        # name
        new.name = self.name
        
        # moderated
        new.moderated = self.moderated
        
        return new
    
    
    def copy_with(self, **keyword_parameters):
        # emoji
        try:
            emoji = keyword_parameters.pop('emoji')
        except KeyError:
            emoji = self.emoji
        else:
            emoji = validate_emoji(emoji)
        
        # name
        try:
            name = keyword_parameters.pop('name')
        except KeyError:
            name = self.name
        else:
            name = validate_name(name)
        
        # moderated
        try:
            moderated = keyword_parameters.pop('moderated')
        except KeyError:
            moderated = self.moderated
        else:
            moderated = validate_moderated(moderated)
        
        
        if keyword_parameters:
            raise TypeError(f'Unused or unsettable attributes: {keyword_parameters!r}.')
        
        
        new = object.__new__(type(self))
        new.id = 0
        new.emoji = emoji
        new.name = name
        new.moderated = moderated
        return new
    
    
    @classmethod
    def precreate(cls, forum_tag_id, **keyword_parameters):
        """
        Precreates a new forum tag instance.
        
        Other Parameters
        ----------------
        emoji : ``None | Emoji``, Optional (Keyword only)
            The tag's emoji.
        
        name : `str`, Optional (Keyword only)
            The tag's name.
        
        moderated : `bool`, Optional (Keyword only)
            Whether this tag can only be added or removed by a user with `manage_threads` permission.
        
        Returns
        -------
        forum_tag : ``ForumTag``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        forum_tag_id = validate_id(forum_tag_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = FORUM_TAGS[forum_tag_id]
        except KeyError:
            self = cls._create_empty(forum_tag_id)
            FORUM_TAGS[forum_tag_id] = self
            
            # Cannot detect when we are really-partial in our channel's scope, so lets assign variables only if we
            # were just created.
            if (processed is not None):
                for item in processed:
                    setattr(self, *item)
        
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new forum tag from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Forum tag data.
        
        Returns
        -------
        self : ``ForumTag``
        """
        forum_tag_id = parse_id(data)
        
        try:
            self = FORUM_TAGS[forum_tag_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = forum_tag_id
            FORUM_TAGS[forum_tag_id] = self
        
        self._update_attributes(data)
        return self
    
    
    def _update_attributes(self, data):
        """
        Updates the forum tag with the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Forum tag data.
        """
        # id
        # Internal field
        
        # emoji
        self.emoji = parse_emoji(data)
        
        # name
        self.name = parse_name(data)
        
        # moderated
        self.moderated = parse_moderated(data)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the forum tag with the given data and returns the changed attributes in an `attribute name - old value`
        relation.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Forum tag data.
        
        Returns
        -------
        old_attributes : `dict<str, object>`
            The updated attributes.
            
            Every item in the dictionary is optional.
            
            +-----------+-------------------+
            | Keys      | Values            |
            +===========+===================+
            | emoji     | ``None | Emoji``  |
            +-----------+-------------------+
            | name      | `str`             |
            +-----------+-------------------+
            | moderated | `bool`            |
            +-----------+-------------------+
        """
        old_attributes = {}
        
        # id
        # Internal field
        
        # emoji
        emoji = parse_emoji(data)
        if (emoji is not self.emoji):
            old_attributes['emoji'] = self.emoji
            self.emoji = emoji
        
        # name
        name = parse_name(data)
        if (name != self.name):
            old_attributes['name'] = self.name
            self.name = name
        
        # moderated
        moderated = parse_moderated(data)
        if (moderated != self.moderated):
            old_attributes['moderated'] = self.moderated
            self.moderated = moderated
        
        return old_attributes
    
    
    @classmethod
    def _create_or_difference_update(cls, data):
        """
        Creates a forum tag. if it already exists calls ``._difference_update_attributes`` on it.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Forum tag data.
        
        Returns
        -------
        forum_tag_and_difference : ``ForumTag``, (`None`, `dict<str, object>`)
        """
        forum_tag_id = parse_id(data)
        
        try:
            self = FORUM_TAGS[forum_tag_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = forum_tag_id
            self._update_attributes(data)
            FORUM_TAGS[forum_tag_id] = self
        
            old_attributes = None
        
        else:
            old_attributes = self._difference_update_attributes(data)
        
        return self, old_attributes
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the forum tag to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether we want to include identifiers as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        # id
        if include_internals:
            put_id(self.id, data, defaults)
        
        # emoji
        put_emoji(self.emoji, data, defaults)
        
        # name
        put_name(self.name, data, defaults)
        
        # moderated
        put_moderated(self.moderated, data, defaults)
        
        return data
    
    
    @property
    def partial(self):
        """
        Returns whether the forum tag is partial.
        
        Returns
        -------
        partial : `bool`
        """
        if self.id == 0:
            return True
        
        return False



__all__ = ('CommandBaseApplicationCommand',)

from scarletio import copy_docs

from .....discord.application_command import (
    ApplicationCommand, ApplicationCommandHandlerType, ApplicationCommandTargetType
)
from .....discord.application_command.application_command_permission.constants import (
    PERMISSION_OVERWRITES_MAX as APPLICATION_COMMAND_PERMISSION_OVERWRITES_MAX
)
from .....discord.user.user.helpers import _try_get_guild_id

from ...utils import SYNC_ID_GLOBAL, SYNC_ID_NON_GLOBAL, UNLOADING_BEHAVIOUR_DELETE, UNLOADING_BEHAVIOUR_INHERIT

from ..command_base import CommandBase


class CommandBaseApplicationCommand(CommandBase):
    """
    Base class for ``Slasher``'s application commands.
    
    Attributes
    ----------
    _exception_handlers : `None`, `list` of `CoroutineFunction`
        Exception handlers added with ``.error`` to the interaction handler.
    
    _parent_reference : `None | WeakReferer<SelfReferenceInterface>`
        Reference to the slasher application command's parent.
    
    _permission_overwrites : `None | dict<int, list<ApplicationCommandPermissionOverwrite>>`
        Permission overwrites applied to the slash command.

    _registered_application_command_ids : `None | dict<int, int>`
        The registered application command ids, which are matched by the command's schema.
        
        If empty set as `None`, if not then the keys are the respective guild's id and the values are the application
        command id.
    
    _schema : `None | ApplicationCommand`
        Internal slot used by the ``.get_schema`` method.
    
    _unloading_behaviour : `int`
        Behaviour what describes what should happen when the command is removed from the slasher.
        
        Can be any of the following:
        
        +-------------------------------+-------+
        | Respective name               | Value |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_DELETE    | 0     |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_KEEP      | 1     |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_INHERIT   | 2     |
        +-------------------------------+-------+
    
    global_ : `bool`
        Whether the command is a global command.
        
        Global commands have their``.guild_ids`` set as `None`.
    
    guild_ids : `None | set<int>`
        The ``Guild``'s id to which the command is bound to.
    
    integration_context_types : `None | tuple<ApplicationCommandIntegrationContextType>`
        The places where the application command shows up.
    
    integration_types : `None | tuple<ApplicationIntegrationType>`
        The options where the application command can be integrated to.
    
    name : `str`
        Application command name. It's length can be in range [1:32].
    
    nsfw : `bool`
        Whether the application command is only allowed in nsfw channels.
    
    required_permissions : ``Permission``
        The required permissions to use the application command inside of a guild.
    """
    __slots__ = (
        '_permission_overwrites', '_registered_application_command_ids', '_schema', '_unloading_behaviour',
        'global_', 'guild_ids', 'integration_context_types', 'integration_types', 'nsfw', 'required_permissions'
    )
    
    @copy_docs(CommandBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        repr_parts = CommandBase._put_repr_parts_into(self, repr_parts)
        
        repr_parts.append(', type = ')
        guild_ids = self.guild_ids
        if guild_ids is None:
            if self.global_:
                type_name = 'global'
            else:
                type_name = 'non-global'
        else:
            type_name = 'guild bound'
        
        repr_parts.append(type_name)
        
        # integration_context_types
        integration_context_types = self.integration_context_types
        if (integration_context_types is not None):
            repr_parts.append(', integration_context_types = ')
            repr_parts.append(repr(integration_context_types))
        
        # integration_types
        integration_types = self.integration_types
        if (integration_types is not None):
            repr_parts.append(', integration_types = ')
            repr_parts.append(repr(integration_types))
        
        # nsfw
        nsfw = self.nsfw
        if (nsfw is not None):
            repr_parts.append(', nsfw = ')
            repr_parts.append(repr(nsfw))
        
        # required_permissions
        required_permissions = self.required_permissions
        if required_permissions:
            repr_parts.append(', required_permissions = ')
            repr_parts.append(repr(required_permissions))
        
        # _unloading_behaviour
        unloading_behaviour = self._unloading_behaviour
        if unloading_behaviour != UNLOADING_BEHAVIOUR_INHERIT:
            repr_parts.append(', unloading_behaviour = ')
            if unloading_behaviour == UNLOADING_BEHAVIOUR_DELETE:
                unloading_behaviour_name = 'delete'
            else:
                unloading_behaviour_name = 'keep'
            
            repr_parts.append(unloading_behaviour_name)
        
        # guild_ids
        if (guild_ids is not None):
            repr_parts.append(', guild_ids = ')
            repr_parts.append(repr(guild_ids))
        
        # _permission_overwrites
        permission_overwrites = self._permission_overwrites
        if (permission_overwrites is not None):
            repr_parts.append(', permission_overwrites = ')
            repr_parts.append(repr(permission_overwrites))
        
        return repr_parts
    
    
    @copy_docs(CommandBase.__hash__)
    def __hash__(self):
        hash_value = CommandBase.__hash__(self)
        
        # _permission_overwrites
        permission_overwrites = self._permission_overwrites
        if (permission_overwrites is not None):
            
            hash_value ^= len(permission_overwrites) << 8
            
            for permission_overwrite in permission_overwrites:
                hash_value ^= hash(permission_overwrite)
        
        # _registered_application_command_ids
        # Internal Field
        
        # _unloading_behaviour
        hash_value ^= (self._unloading_behaviour + 1) << 12
        
        # integration_context_types
        integration_context_types = self.integration_context_types
        if (integration_context_types is not None):
            hash_value ^= len(integration_context_types) << 9
            for integration_context_type in integration_context_types:
                hash_value ^= integration_context_type.value << 13
        
        # integration_types
        integration_types = self.integration_types
        if (integration_types is not None):
            hash_value ^= len(integration_types) << 7
            for integration_type in integration_types:
                hash_value ^= integration_type.value << 11
        
        # guild_ids
        guild_ids = self.guild_ids
        if (guild_ids is not None):
            hash_value ^= len(guild_ids) << 18
            
            for guild_id in guild_ids:
                hash_value ^= guild_id
        
        # nsfw
        nsfw = self.nsfw
        if (nsfw is not None):
            hash_value ^ nsfw << 22
        
        # required_permissions
        hash_value ^= self.required_permissions
        
        return hash_value
    
    
    @copy_docs(CommandBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not CommandBase._is_equal_same_type(self, other):
            return False
        
        # _permission_overwrites
        if self._permission_overwrites != other._permission_overwrites:
            return False
        
        # _registered_application_command_ids
        # Internal field

        # _unloading_behaviour
        if self._unloading_behaviour != other._unloading_behaviour:
            return False
        
        # integration_context_types
        if self.integration_context_types != other.integration_context_types:
            return False
        
        # integration_types
        if self.integration_types != other.integration_types:
            return False
        
        # guild_ids
        if self.guild_ids != other.guild_ids:
            return False
        
        # global
        if self.global_ != other.global_:
            return False
        
        # nsfw
        if self.nsfw != other.nsfw:
            return False
        
        # required_permissions
        if self.required_permissions != other.required_permissions:
            return False
        
        return True
    
    
    @copy_docs(CommandBase.copy)
    def copy(self):
        new = CommandBase.copy(self)
        
        # _permission_overwrites
        permission_overwrites = self._permission_overwrites
        if (permission_overwrites is not None):
            permission_overwrites = {
                guild_id: permission_overwrite.copy() for
                guild_id, permission_overwrite in permission_overwrites.items()
            }
        new._permission_overwrites = permission_overwrites
        
        # _registered_application_command_ids
        new._registered_application_command_ids = None
        
        # _schema
        new._schema = None
        
        # _unloading_behaviour
        new._unloading_behaviour = self._unloading_behaviour
        
        # integration_context_types
        integration_context_types = self.integration_context_types
        if (integration_context_types is not None):
            integration_context_types = (*integration_context_types,)
        new.integration_context_types = integration_context_types
        
        # integration_types
        integration_types = self.integration_types
        if (integration_types is not None):
            integration_types = (*integration_types,)
        new.integration_types = integration_types
        
        # guild_ids
        guild_ids = self.guild_ids
        if (guild_ids is not None):
            guild_ids = guild_ids.copy()
        new.guild_ids = guild_ids
        
        # global_
        new.global_ = self.global_
        
        # nsfw
        new.nsfw = self.nsfw
        
        # required_permissions
        new.required_permissions = self.required_permissions
        
        return new
    
    
    @property
    def target_type(self):
        """
        Returns command's target type.
        
        This property is a placeholder for subclasses which actually implement it.
        
        Returns
        -------
        target ``ApplicationCommandTargetType``
        """
        return ApplicationCommandTargetType.none
    
    
    @target_type.setter
    def target_type(self, value):
        pass
    
    
    @property
    def handler_type(self):
        """
        Returns command's handler type.
        
        This property is a placeholder for subclasses which actually implement it.
        
        Returns
        -------
        handler :  ``ApplicationCommandHandlerType``
        """
        return ApplicationCommandHandlerType.none
    
    
    @handler_type.setter
    def handler_type(self, value):
        pass
    
    
    @property
    def default(self):
        """
        Returns the command's default.
        
        This property is a placeholder for subclasses which actually implement it.
        
        Returns
        -------
        default `bool`
        """
        return False
    
    
    @default.setter
    def default(self, value):
        return
    
    
    @property
    def description(self):
        """
        Returns the command's description.
        
        This property is a placeholder for subclasses which actually implement it.
        
        Returns
        -------
        description `None | str`
        """
        return None
    
    
    @description.setter
    def description(self, value):
        return
    
    
    async def invoke_auto_completion(self, client, interaction_event, auto_complete_option):
        """
        Calls the auto completion function of the slasher application command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        
        interaction_event : ``InteractionEvent``
            The received interaction event.
        
        auto_complete_option : `InteractionMetadataApplicationCommandAutocomplete | InteractionOption`
            The option to autocomplete.
        """
        pass
    
    # ---- Mention ----
    
    @property
    @copy_docs(CommandBase.mention)
    def mention(self):
        return self._build_mention(self._get_application_id_for_mention(), None)
    
    
    @copy_docs(CommandBase.mention_at)
    def mention_at(self, guild):
        return self._build_mention(self._get_application_id_for_mention_at(guild), None)
    
    
    def _mention_recursive(self, *sub_command_names):
        """
        Returns the application command's mention.
        
        Parameters
        ----------
        *sub_command_names : `str`
            Already included sub-command name stack to mention.
        
        Returns
        -------
        mention : `str`
        """
        return self._build_mention(self._get_application_id_for_mention(), sub_command_names)
    
    
    def _mention_at_recursive(self, guild, *sub_command_names):
        """
        Returns the application command's mention.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild to mention the command at.
        
        *sub_command_names : `str`
            Already included sub-command name stack to mention.
        
        Returns
        -------
        mention : `str`
        """
        return self._build_mention(self._get_application_id_for_mention_at(guild), sub_command_names)
    
    
    def _get_application_id_for_mention(self):
        """
        Gets the application command's identifier to mention.
        
        This function is used by ``.mention`` and works if the application command is global or if it is added only
        to one guild.
        
        Returns
        -------
        application_command_id : `int`
        """
        registered_application_command_ids = self._registered_application_command_ids
        if (registered_application_command_ids is None) or (len(registered_application_command_ids) > 1):
            application_command_id = 0
        else:
            try:
                application_command_id = registered_application_command_ids[SYNC_ID_GLOBAL]
            except KeyError:
                application_command_id = next(iter(registered_application_command_ids.values()))
        
        return application_command_id
    
    
    def _get_application_id_for_mention_at(self, guild):
        """
        Gets the application command's identifier to mention at the specified guild.
        
        This function is used by ``.mention_at``.
        
        Parameters
        ----------
        guild : ``Guild | int`
            The guild to mention the application command at.
        
        Returns
        -------
        application_command_id : `int`
        """
        registered_application_command_ids = self._registered_application_command_ids
        if registered_application_command_ids is None:
            application_command_id = 0
        else:
            if self.global_:
                application_command_id = registered_application_command_ids.get(SYNC_ID_GLOBAL, 0)
            else:
                application_command_id = registered_application_command_ids.get(_try_get_guild_id(guild), 0)
        
        return application_command_id
    
    
    def _build_mention(self, application_command_id, sub_command_name_stack):
        """
        Builds the application command's mention.
        
        Parameters
        ----------
        application_command_id : `int`
            The respective application command's identifier.
        
        sub_command_name_stack : `None | tuple<str>`
            Additional sub command names to include in the mention.
        
        Returns
        -------
        mention : `str`
        """
        mention_parts = []
        
        if application_command_id:
            mention_parts.append('<')
            
        mention_parts.append('/')
        mention_parts.append(self.name)
        
        if (sub_command_name_stack is not None):
            for sub_command_name in sub_command_name_stack:
                mention_parts.append(' ')
                mention_parts.append(sub_command_name)
        
        if application_command_id:
            mention_parts.append(':')
            mention_parts.append(str(application_command_id))
            mention_parts.append('>')
        
        return ''.join(mention_parts)
    
    # ---- Utility ----
    
    def get_real_command_count(self):
        """
        Gets the real command count of the command. This includes every sub attached to it as well.
        
        Returns
        -------
        real_command_count: `int`
        """
        return 1
    
    
    @property
    def interactions(self):
        """
        Enables you to add sub-commands or sub-categories to the command.
        
        Raises
        ------
        RuntimeError
            Self is not a category.
        """
        raise RuntimeError(
            f'The {type(self).__name__} is not a category.'
        )
    
    
    # ---- Permission overwrites ----

    def add_permission_overwrite(self, guild_id, permission_overwrite):
        """
        Adds an overwrite to the slash command.
        
        Parameters
        ----------
        guild_id : `int`
            The guild's id where the overwrite will be applied.
        
        permission_overwrite : `None | ApplicationCommandPermissionOverwrite`
            The permission overwrite to add.
        
        Raises
        ------
        ValueError
            - Each command in each guild can have up to `10` overwrite, which is already reached.
        """
        permission_overwrites = self._permission_overwrites
        if permission_overwrites is None:
            self._permission_overwrites = permission_overwrites = {}
        
        permission_overwrites_for_guild = permission_overwrites.get(guild_id, None)
        
        if (
            (permission_overwrites_for_guild is not None) and
            (len(permission_overwrites_for_guild) >= APPLICATION_COMMAND_PERMISSION_OVERWRITES_MAX)
        ):
            raise ValueError(
                f'`Each command in each guild can have up to '
                f'{APPLICATION_COMMAND_PERMISSION_OVERWRITES_MAX} permission overwrites which is already reached.'
            )
        
        if (permission_overwrites_for_guild is not None) and (permission_overwrite is not None):
            target_id = permission_overwrite.target_id
            for index in range(len(permission_overwrites_for_guild)):
                iter_permission_overwrites = permission_overwrites_for_guild[index]
                
                if iter_permission_overwrites.target_id != target_id:
                    continue
                
                if permission_overwrite.allow == iter_permission_overwrites.allow:
                    return
                
                del permission_overwrites_for_guild[index]
                
                if permission_overwrites_for_guild:
                    return
                
                permission_overwrites[guild_id] = None
                return
        
        if permission_overwrite is None:
            if permission_overwrites_for_guild is None:
                permission_overwrites[guild_id] = None
        else:
            if permission_overwrites_for_guild is None:
                permission_overwrites[guild_id] = permission_overwrites_for_guild = []
            
            permission_overwrites_for_guild.append(permission_overwrite)
    
    
    def get_permission_overwrites_for(self, guild_id):
        """
        Returns the slash command's permissions overwrites for the given guild.
        
        Returns
        -------
        permission_overwrites : `None | list<ApplicationCommandPermissionOverwrite>`
            Returns `None` instead of an empty list.
        """
        permission_overwrites = self._permission_overwrites
        if (permission_overwrites is not None):
            return permission_overwrites.get(guild_id, None)
    
    # ---- Sync ----
    
    def _get_permission_sync_ids(self):
        """
        Gets the permission overwrite guild id-s which should be synced.
        
        Returns
        -------
        permission_sync_ids : `set<int>`
        """
        permission_sync_ids = set()
        guild_ids = self.guild_ids
        # If the command is guild bound, sync it in every guild, if not, then sync it in every guild where it has an
        # a permission overwrite.
        if (guild_ids is None):
            permission_overwrites = self._permission_overwrites
            if (permission_overwrites is not None):
                permission_sync_ids.update(permission_overwrites.keys())
        else:
            permission_sync_ids.update(guild_ids)
        
        return permission_sync_ids
    
    
    def _register_guild_and_application_command_id(self, guild_id, application_command_id):
        """
        Registers an application command's identifier to the ``SlashCommand`.
        
        Parameters
        ----------
        application_command_id : `int`
            The application command's identifier.
        
        guild_id : `int`
            The guild where the application command is in.
        """
        registered_application_command_ids = self._registered_application_command_ids
        if registered_application_command_ids is None:
            registered_application_command_ids = self._registered_application_command_ids = {}
        
        registered_application_command_ids[guild_id] = application_command_id
    
    
    def _unregister_guild_and_application_command_id(self, guild_id, application_command_id):
        """
        Unregisters an application command's identifier from the ``SlashCommand`.
        
        Parameters
        ----------
        guild_id : `int`
            The guild's id, where the application command is in.
        
        application_command_id : `int`
            The application command's identifier.
        """
        registered_application_command_ids = self._registered_application_command_ids
        if registered_application_command_ids is not None:
            try:
                maybe_application_command_id = registered_application_command_ids[guild_id]
            except KeyError:
                pass
            else:
                if maybe_application_command_id == application_command_id:
                    del registered_application_command_ids[guild_id]
                    
                    if not registered_application_command_ids:
                        self._registered_application_command_ids = None
    
    
    def _pop_application_command_id_for(self, guild_id):
        """
        Pops the given application command id from the command for the respective guild.
        
        Parameters
        ----------
        guild_id : `int`
            A guild's identifier.
        
        Returns
        -------
        application_command_id : `int`
            The popped application command's identifier. Returns `0` if nothing is matched.
        """
        registered_application_command_ids = self._registered_application_command_ids
        if registered_application_command_ids is None:
            application_command_id = 0
        else:
            try:
                application_command_id = registered_application_command_ids.pop(guild_id)
            except KeyError:
                application_command_id = 0
            else:
                if not registered_application_command_ids:
                    self._registered_application_command_ids = None
        
        return application_command_id
    
    
    def _iter_application_command_ids(self):
        """
        Iterates over all the registered application command id-s added to the slash command.
        
        This method is an iterable generator.
        
        Yields
        ------
        application_command_id : `int`
        """
        registered_application_command_ids = self._registered_application_command_ids
        if (registered_application_command_ids is not None):
            yield from registered_application_command_ids.values()
    
    
    def _exhaust_application_command_ids(self):
        """
        Iterates over all the registered application command id-s added to the slash command and removes them.
        
        This method is an iterable generator.
        
        Yields
        ------
        application_command_id : `int`
        """
        registered_application_command_ids = self._registered_application_command_ids
        while (registered_application_command_ids is not None):
            guild_id, application_command_id = registered_application_command_ids.popitem()
            if not registered_application_command_ids:
                registered_application_command_ids = None
                self._registered_application_command_ids = None
            
            yield application_command_id
    
    
    def _iter_sync_ids(self):
        """
        Iterates over all the respective sync ids of the command. If the command is a guild bound command, then will
        iterate over it's guild's id-s.
        
        This method is a generator, what should be used inside of a `for` loop.
        
        Yields
        ------
        sync_id : `int`
        """
        if self.global_:
            yield SYNC_ID_GLOBAL
            return
        
        guild_ids = self.guild_ids
        if guild_ids is None:
            yield SYNC_ID_NON_GLOBAL
            return
        
        yield from guild_ids
    
    
    def _iter_guild_ids(self):
        """
        Iterates over all the guild identifiers used by the command.
        
        Yields
        ------
        guild_id : `int`
        """
        registered_application_command_ids = self._registered_application_command_ids
        if registered_application_command_ids is not None:
            for sync_id in registered_application_command_ids:
                if sync_id > (1 << 22):
                    yield sync_id
    
    # ---- Schema ----

    def get_schema(self):
        """
        Returns an application command schema representing the slash command.
        
        Returns
        -------
        schema : ``ApplicationCommand``
        """
        schema = self._schema
        if schema is None:
            schema = self._schema = self._get_schema()
        
        return schema
    
    
    def _get_schema(self):
        """
        Creates a new application command schema representing the slash command.
        
        Returns
        -------
        schema : ``ApplicationCommand``
        """
        schema = ApplicationCommand(
            self.name,
            self.description,
            handler_type = self.handler_type,
            integration_context_types = self.integration_context_types,
            integration_types = self.integration_types,
            options = self._get_schema_options(),
            nsfw = self.nsfw,
            required_permissions = self.required_permissions,
            target_type = self.target_type,
        )
        
        parent_reference = self._parent_reference
        if (parent_reference is not None):
            parent = parent_reference()
            if (parent is not None):
                schema = schema.with_translation(parent._translation_table)
        
        return schema
    
    
    def _get_schema_options(self):
        """
        Gets the schema options for the application command.
        
        Returns
        -------
        application_command_options : `None | list<ApplicationCommandOption>`
        """
        return None



__all__ = ('InvalidToken',)


INVALID_TOKEN_ERROR_MESSAGE = 'Invalid token, please update it, then start the client again.'


class InvalidToken(BaseException):
    __slots__ = ()
    
    def __new__(cls):
        return BaseException.__new__(cls, INVALID_TOKEN_ERROR_MESSAGE)
    
    __init__ = object.__init__
